const fs = require('fs');
const { ColorVariant } = require('./theme-color');
const { Parser } = require('./theme-parser');

/**
 * @typedef {object} Theme
 * @property {string} prefix
 * @property {Record<string, string>} readonly
 * @property {Record<string, string>} public
 * @property {Record<string, string>} colors
 * @property {Record<string, string>} icons
 */
/**
 * @typedef {object} ColorVariantInfo
 * @property {string} name      color name
 * @property {string} variant   variant name
 * @property {string} value     color value (hex,rgb,hsl etc)
 */
/**
 * @typedef {object} ColorVariant
 * @property {string} name      color name
 * @property {string} value     color value (hex,rgb,hsl etc)
 */
/**
 * @typedef {object} ColorInfo
 * @property {string} name
 * @property {string[]?} variants
 */

/**
 * Creates a color variant
 * @param {ColorVariantInfo} colorInfo
 * @returns {ColorVariant}
 */
const createColor = ({ name, variant, value }) => ({ name: `${name}-${variant}`, value: ColorVariant[variant](value) });

/**
 * Defines colors
 * @param {Record<string, ColorInfo>} colors
 * @returns {Record<string, string>}
 */
const defineColors = colors => {
    return Object.entries(colors).reduce((acc, [name, info]) => {
        acc[name] = info.value;
        (info.variants || []).forEach(variant => {
            const color = createColor({ name, variant, value: info.value });
            acc[color.name] = color.value;
        });
        return acc;
    }, {});
};

/**
 * Defines icons
 * @param {Record<string, string>} icons
 * @returns {Record<string, string>}
 */
const defineIcons = icons =>
    Object.entries(icons).reduce(
        (acc, [key, val]) => ({ ...acc, [key]: `url('data:image/svg+xml;utf8,${val.replace(/#/g, '%23')}')` }),
        {}
    );

const getCssVar = name => `.var(${name})[@return]`;

const buildCssVar = (prefix, name, value) => `--${prefix}${name}: ${value};`;

const buildLessVar = (name, value) => `@${name}: ${value};`;

const mapObj = (obj, fn) => Object.entries(obj).map(args => fn(...args));

/**
 * @param {Theme} theme
 * @param {string?} path
 * @returns {string}
 */
const build = ({ prefix, readonly, public, protected, colors, icons }, path = null) => {
    const separator = { css: '\n\t', less: '\n' };
    const out = `
:root {
    // readonly
    ${mapObj(readonly, (name, value) => buildCssVar(prefix, name, value)).join(separator.css)}
    
    // public
    ${mapObj(public, (name, value) => buildCssVar(prefix, name, value)).join(separator.css)}
    
    // protected
    ${mapObj(protected, (name, value) => buildCssVar(prefix, name, value)).join(separator.css)}
    
    // colors
    ${mapObj(colors, (name, value) => buildCssVar(prefix, `color-${name}`, value)).join(separator.css)}

    // icons
    ${mapObj(icons, (name, value) => buildCssVar(prefix, name, value)).join(separator.css)}
}

@prefix: ~'${prefix}';
// readonly
${mapObj(readonly, (name, value) => buildLessVar(name, value)).join(separator.less)}

// public
${mapObj(public, name => buildLessVar(name, getCssVar(name))).join(separator.less)}

// protected
${mapObj(protected, name => buildLessVar(name, getCssVar(name))).join(separator.less)}

// colors
${mapObj(colors, name => buildLessVar(`color-${name}`, getCssVar(`color-${name}`))).join(separator.less)}

// icons
${mapObj(icons, name => buildLessVar(name, getCssVar(name))).join(separator.less)}
`;

    if (path) {
        fs.writeFileSync(path, out, { encoding: 'utf-8' });
    }
    return out;
};

module.exports = { build, createColor, defineColors, defineIcons, Parser };
